import shutil
import subprocess
import tempfile

from flask import make_response

from steady_cell_phenotype._util import *


def compute_fixed_points(knockout_model, variables, continuous):
    """ Run the fixed-point finding computation """

    # check to make sure that we have macaulay installed
    macaulay_executable = shutil.which('M2')
    if macaulay_executable is None:
        return make_response(error_report("Macaulay2 was not found on the server; we cannot do as you ask."))

    # we are set to do the computation
    with tempfile.TemporaryDirectory() as tmp_dir_name:
        with open(tmp_dir_name + '/model.txt', 'w') as model_file:
            model_file.write(knockout_model)
            model_file.write('\n')

        # convert the model to polynomials
        non_continuous_vars = [variable for variable in variables if not continuous[variable]]
        if len(non_continuous_vars) > 0:
            continuity_params = ['-c', '-comit'] + [variable for variable in variables if not continuous[variable]]
        else:
            continuity_params = ['-c']

        convert_to_poly_process = \
            subprocess.run([get_resource_path('scp_converter.py'), '-n',
                            '-i', tmp_dir_name + '/model.txt',
                            '-o', tmp_dir_name + '/model-polys.txt'] + continuity_params,
                           capture_output=True)

        if convert_to_poly_process.returncode != 0:
            return make_response(error_report(
                'Error running converter!\n{}\n{}'.format(html_encode(convert_to_poly_process.stdout),
                                                          html_encode(convert_to_poly_process.stderr))))

        template_contents = get_text_resource('find_steady_states.m2-template')
        with open(tmp_dir_name + '/find_steady_states.m2', 'w') as macaulay_script:
            macaulay_script.write(template_contents.format(
                polynomial_file=tmp_dir_name + '/model-polys.txt',
                output_file=tmp_dir_name + '/results.txt'))

        # TODO: put a limit on the amount of time this can run
        find_steady_states_process = \
            subprocess.run([macaulay_executable,
                            '--script',
                            tmp_dir_name + '/find_steady_states.m2'],
                           capture_output=True)
        if find_steady_states_process.returncode != 0:
            with open(tmp_dir_name + '/model-polys.txt', 'r') as poly_file:
                return make_response(error_report(
                    'Error running Macaulay!\n<br>\n{}\n<br>\n{}\n<br>\n{}'.format(
                        html_encode(find_steady_states_process.stdout),
                        html_encode(find_steady_states_process.stderr),
                        html_encode(poly_file.read()))))

        def process_line(line):
            line = line.strip()
            if len(line) < 2:  # should at least have {}
                return None
            elif line[0] != '{' or line[-1] != '}':
                raise RuntimeError("Malformed response from Macaulay")

            line = line[1:-1]  # strip the {}'s
            try:
                results = map(int, line.split(','))
            except ValueError:
                raise RuntimeError("Malformed response from Macaulay")

            # make sure they are in 0,1,2
            results = map(lambda n: (n + 3) % 3, results)

            return tuple(results)

        try:
            with open(tmp_dir_name + '/results.txt', 'r') as file:
                fixed_points = [process_line(line) for line in file]
        except (IOError, RuntimeError):
            return make_response(error_report('Malformed response from Macaulay!'))

    # respond with the results-of-computation page
    return make_response(
        render_template('compute-fixed-points.html', variables=variables, fixed_points=fixed_points))



import html
import importlib.resources
import itertools
import pathlib
import re
import string
from typing import Callable, Dict, Iterator, List, Tuple, Union

from attr import attrib, attrs
from flask import render_template
import numba
import numpy as np

from steady_cell_phenotype.equation_system import EquationSystem

MAX_SUPPORTED_VARIABLE_STATES = 6


def get_resource_path(filename: str) -> pathlib.Path:
    """
    Obtain the location of steady_cell_phenotype's data files regardless of package location.

    Parameters
    ----------
    filename: str
        File to obtain
    Returns
    -------
    File's contents as a string
    """
    # See: https://docs.python.org/3.7/library/importlib.html#module-importlib.resources
    with importlib.resources.path("steady_cell_phenotype", filename) as path:
        return path


def get_text_resource(filename: str) -> str:
    """
    Obtain the contents of steady_cell_phenotype's data files regardless of package location.

    Parameters
    ----------
    filename: str
        File to obtain
    Returns
    -------
    File's contents as a string
    """
    # See: https://docs.python.org/3.7/library/importlib.html#module-importlib.resources
    with importlib.resources.open_text(
            "steady_cell_phenotype", filename
            ) as file_handle:
        data = file_handle.read()
    return data


def html_encode(msg):
    if type(msg) is not str:
        msg = msg.decode()
    return html.escape(msg).replace("\n", "<br>").replace(" ", "&nbsp")


def error_report(error_string):
    """ display error reports from invalid user input """
    return render_template("error.html", error_message=error_string)


def get_model_variables(model) -> Tuple[List[str], List[str]]:
    variables: List[str] = []
    right_sides: List[str] = []
    too_many_eq_msg = "Count of ='s on line {lineno} was {eq_count} but each line must have a single = sign."
    zero_len_var_msg = "No variable found before = on line {lineno}."
    zero_len_rhs_msg = "No right hand side of equation on line {lineno}."
    invalid_var_name_msg = (
        "One line {lineno}, variable name must be alpha-numeric (plus underscore), "
        "and begin with a letter.")
    for lineno, line in enumerate(model.splitlines(), start=1):
        # check for _one_ equals sign
        if line.count("=") != 1:
            raise Exception(
                    too_many_eq_msg.format(lineno=lineno, eq_count=line.count("="))
                    )
        variable: str
        rhs: str
        variable, rhs = line.split("=")
        variable = variable.strip()
        rhs = rhs.strip()
        # check to see if lhs is a valid symbol. TODO: what variable names does scp_converter.py allow?
        if len(variable) == 0:
            raise Exception(zero_len_var_msg.format(lineno=lineno))
        if (not re.match(r"^\w+$", variable)
                or variable[0] not in string.ascii_letters):  # TODO: 2nd test may be redundant
            raise Exception(invalid_var_name_msg.format(lineno=lineno))
        variables.append(variable)
        # do _minimal_ checking on RHS
        if len(rhs) == 0:
            raise Exception(zero_len_rhs_msg.format(lineno=lineno))
        right_sides.append(rhs)
    return variables, right_sides


def decode_int(coded_value, num_variables):
    """ Decode long-form int into trinary """
    if isinstance(coded_value, str):
        if coded_value[:2] == "0x":
            coded_value = int(coded_value, 16)
        else:
            coded_value = int(coded_value)
    exploded_values = []
    for _ in range(num_variables):
        next_value = coded_value % 3
        exploded_values.append(next_value)
        coded_value = (coded_value - next_value) // 3
    exploded_values.reverse()
    return exploded_values


@attrs
class StreamingStats(object):
    """
    Implements Welford's online algorithm for mean and variance calculation. See Knuth Vol 2, pg 232
    """

    mean: float = attrib(init=False, default=float("nan"))
    scaled_var: float = attrib(init=False, default=float("nan"))
    var: float = attrib(init=False, default=float("nan"))
    count: int = attrib(init=False, default=0)

    def add(self, datum: float):
        self.count += 1
        if self.count == 1:
            self.mean = datum
            self.scaled_var = 0
            self.var = 0
        else:
            old_mean = self.mean
            self.mean += (datum - old_mean) / self.count
            self.scaled_var += (datum - old_mean) * (datum - self.mean)

        if self.count > 1:
            self.var = self.scaled_var / (self.count - 1)
        else:
            self.var = 0.0


@attrs
class BinCounter(object):
    """
    Utility class for streaming bincounts.
    """

    bins: np.ndarray = attrib(init=False, default=np.zeros(0, dtype=int))
    max: int = attrib(init=False, default=-1)

    def total(self) -> int:
        return int(np.sum(self.bins))

    def add(self, datum: Union[int, "BinCounter"]):
        if isinstance(datum, int):
            if datum >= len(self.bins):
                old_bins = self.bins
                self.bins = np.zeros(1 + int(1.5 * datum), dtype=int)
                self.bins[: len(old_bins)] = old_bins
            self.bins[datum] += 1
            self.max = max(self.max, datum)
        elif isinstance(datum, BinCounter):
            new_max = max(self.max, datum.max)
            new_bins = np.zeros(shape=(new_max + 1))
            new_bins[: self.max + 1] = self.bins[: self.max + 1]
            new_bins[: datum.max + 1] += datum.bins[: datum.max + 1]
            self.max = new_max
            self.bins = new_bins
        else:
            raise RuntimeError("Cannot handle this.")

    def trimmed_bins(self):
        return np.trim_zeros(filt=self.bins, trim="b")


@numba.njit
def ternary_hash(arr: np.ndarray) -> int:
    accumulator: int = 0
    for idx in range(len(arr)):
        accumulator = 3 * accumulator + arr[idx]
    return int(accumulator)


class HashableNdArray(object):
    """
    A wrapper to make numpy based state arrays hashable
    """

    array: np.ndarray
    hash: int

    def __init__(self, array):
        self.array = array
        # view the array as a ternary expansion; a bijective hash
        self.hash = ternary_hash(array)

    def __repr__(self):
        # return 'HashableNdArray(array=' + repr(self.array) + ')'
        return repr(self.array)

    def __str__(self):
        return str(self.array)

    def __getitem__(self, item):
        return self.array[item]

    def __eq__(self, other: "HashableNdArray"):
        # decided on array_equiv rather than array_equal, because I don't want to worry about
        # the case where the shape is (1,n) or (n,1) and being compared to (n,) or some
        # other such nonsense.
        #
        # note: as this is a performance critical point, I am not going to check the type of 'other'
        # and instead trust that I don't do anything brain-dead.
        return self.hash == other.hash and np.array_equiv(self.array, other.array)

    def __hash__(self):
        return self.hash


def complete_search_generator(
        variables: List[str], constants_vals: Dict[str, int]
        ) -> Iterator[np.ndarray]:
    """
    Generator which yields all possible states, with constant variables fixed

    Returns
    -------
    Iterator[np.ndarray]
    """
    constants = tuple(constants_vals.keys())
    num_non_constant_vars = len(variables) - len(constants)
    non_constant_vars = list(set(variables) - set(constants))
    for var_dict in map(
            lambda tup: dict(zip(tup[0], tup[1])),
            zip(
                    itertools.repeat(non_constant_vars),
                    itertools.product(range(3), repeat=num_non_constant_vars),
                    ),
            ):
        state = np.fromiter(
                (
                    var_dict[var] if var in var_dict else constants_vals[var]
                    for var in variables
                    ),
                dtype=np.int64,
                )
        yield state


def random_search_generator(
        num_iterations,
        variables: List[str],
        constants_vals: Dict[str, int],
        batch_size=2000,
        ) -> Iterator[np.ndarray]:
    """
    Generates `num_iterations` random states, with constant variables fixed.

    Parameters
    ----------
    num_iterations : int
        number of iterations to generate
    variables : List[str]
        ordered list of variables
    constants_vals : Dict[str, int]
        dictionary of variables with their constant iteration
    batch_size : int
        size of internal batch, for performance tuning, no other external effect

    Returns
    -------
    Iterator[np.ndarray]
    """
    num_variables = len(variables)
    constant_indices = np.array(
            [idx for idx, var in enumerate(variables) if var in constants_vals],
            dtype=np.int64,
            )
    constant_val_arr = np.array(
            [constants_vals[var] for var in variables if var in constants_vals],
            dtype=np.int64,
            )
    count = 0
    while count < num_iterations:
        init_states = np.random.randint(3, size=(batch_size, num_variables), dtype=int)
        # set constant values
        init_states[:, constant_indices] = constant_val_arr

        for idx in range(batch_size):
            yield init_states[idx, :]
            count += 1
            if count >= num_iterations:
                return


def batcher(
        state_gen: Iterator[np.ndarray], variables, batch_size
        ) -> Iterator[np.ndarray]:
    num_variables = len(variables)
    batch: np.ndarray
    try:
        while True:
            batch = np.zeros(shape=(batch_size, num_variables), dtype=int)
            for idx in range(batch_size):
                batch[idx] = next(state_gen)
            yield batch
    except StopIteration:
        # noinspection PyUnboundLocalVariable
        yield batch[:idx, :]


def get_phased_trajectory(init_state: np.ndarray,
                          update_fn: Callable) -> Tuple[np.ndarray, HashableNdArray]:
    """
    evolve an initial state until it reaches a limit cycle

    Parameters
    ----------
    init_state
    update_fn

    Returns
    -------
    trajectory, phase-point pair
    """
    state = init_state
    trajectory = list()
    trajectory_set = set()  # set lookup should be faster

    # compute state by state until we have a repeat
    hashable_state = HashableNdArray(state)
    while hashable_state not in trajectory_set:
        trajectory.append(hashable_state)
        trajectory_set.add(hashable_state)
        state = update_fn(state)
        hashable_state = HashableNdArray(state)

    # separate trajectory into in-bound and limit-cycle parts
    repeated_state = HashableNdArray(state)
    repeated_state_index = trajectory.index(repeated_state)
    limit_cycle = trajectory[repeated_state_index:]

    # find state in limit cycle with smallest hash (i.e. smallest lexicographic
    # ordering if there is no integer overflow)
    # this is our phase fixing point
    cycle_min_index: int = 0
    cycle_min: int = hash(limit_cycle[0])
    for idx in range(1, len(limit_cycle)):
        nxt_hash: int = hash(limit_cycle[idx])
        if nxt_hash < cycle_min:
            cycle_min_index = idx
            cycle_min = nxt_hash

    # get trajectory with phase
    phase_idx: int = len(trajectory) - len(limit_cycle) + cycle_min_index
    phased_trajectory = np.array(
            [hashable.array for hashable in trajectory[:phase_idx]], dtype=np.int64
            )

    return phased_trajectory, trajectory[phase_idx]


def get_trajectory(init_state: np.ndarray,
                   update_fn: Callable) -> Tuple[np.ndarray, np.ndarray]:
    """
    evolve an initial state until it reaches a limit cycle

    Parameters
    ----------
    init_state
    update_fn

    Returns
    -------
    trajectory, limit cycle
    """
    state = init_state
    trajectory = list()
    trajectory_set = set()  # set lookup should be faster

    # compute state by state until we have a repeat
    hashable_state = HashableNdArray(state)
    while hashable_state not in trajectory_set:
        trajectory.append(hashable_state)
        trajectory_set.add(hashable_state)
        state = update_fn(state)
        hashable_state = HashableNdArray(state)

    # separate trajectory into in-bound and limit-cycle parts
    repeated_state = HashableNdArray(state)
    repeated_state_index = trajectory.index(repeated_state)

    trimmed_trajectory = np.array([hashable.array for hashable in trajectory[:repeated_state_index]])
    limit_cycle = np.array([hashable.array for hashable in trajectory[repeated_state_index:]])

    return trimmed_trajectory, limit_cycle


def process_model_text(model_text: str,
                       knockouts: Dict[str, int],
                       continuous: Dict[str, bool]
                       ) -> Tuple[List[str], Callable, EquationSystem]:
    equation_system = EquationSystem.from_text(model_text)
    equation_system = equation_system.continuous_functional_system(
            continuous_vars=tuple([var for var in continuous if continuous[var]])
            )
    equation_system = equation_system.knockout_system(knockouts)

    # create an update function
    variables: List[str]
    update_fn: Callable
    variables, update_fn = equation_system.as_numpy()

    return variables, update_fn, equation_system



from collections import defaultdict
import functools
import itertools
import math
from math import ceil
import tempfile
from typing import Dict, Iterator, List, NamedTuple, Tuple

from flask import make_response, Markup, render_template
import matplotlib
import matplotlib.pyplot as plt
import numba
import numpy as np
import pathos

from steady_cell_phenotype._util import (
    batcher,
    BinCounter,
    complete_search_generator,
    get_phased_trajectory,
    HashableNdArray,
    process_model_text,
    random_search_generator,
    )

matplotlib.use("agg")


class TrajectoryStatistics(NamedTuple):
    trajectory_length_counts: Dict[HashableNdArray, BinCounter]
    data_counts: Dict[HashableNdArray, np.ndarray]
    means: Dict[HashableNdArray, np.ndarray]
    variances: Dict[HashableNdArray, np.ndarray]

    @classmethod
    def make(cls, num_variables: int):
        return TrajectoryStatistics(
                trajectory_length_counts=defaultdict(lambda: BinCounter()),
                data_counts=defaultdict(lambda: np.zeros(shape=0, dtype=int)),
                means=defaultdict(lambda: np.zeros(shape=(0, num_variables))),
                variances=defaultdict(lambda: np.zeros(shape=(0, num_variables))),
                )


def batch_trajectory_process(batch, update_fn) -> TrajectoryStatistics:
    num_samples = batch.shape[0]
    num_variables = batch.shape[1]

    trajectory_statistics = TrajectoryStatistics.make(num_variables)
    # unpack
    trajectory_length_counts = trajectory_statistics.trajectory_length_counts
    data_counts = trajectory_statistics.data_counts
    means = trajectory_statistics.means
    scaled_variances = trajectory_statistics.variances

    for batch_idx in range(num_samples):
        trajectory, phase_state = get_phased_trajectory(batch[batch_idx, :], update_fn)
        trajectory_length_counts[phase_state].add(len(trajectory))

        # extract arrays, to reduce dict lookup again
        data_count = data_counts[phase_state]
        mean = means[phase_state]
        scaled_variance = scaled_variances[phase_state]

        old_len: int = data_count.shape[0]
        trajectory_len: int = trajectory.shape[0]

        if trajectory_len == 0:
            continue

        # resize, if necessary
        if old_len < trajectory_len:
            # extend data_count
            new_data_count: np.ndarray = np.zeros(shape=trajectory_len, dtype=int)
            new_data_count[:old_len] = data_count
            data_count: np.ndarray = new_data_count
            data_counts[phase_state] = new_data_count
            # extend means
            new_mean: np.ndarray = np.zeros(shape=(trajectory_len, num_variables))
            new_mean[:old_len, :] = mean
            mean = new_mean
            means[phase_state] = new_mean
            # extend scaled variance
            new_scaled_variance: np.ndarray = np.zeros(
                    shape=(trajectory_len, num_variables)
                    )
            new_scaled_variance[:old_len] = scaled_variance
            scaled_variance = new_scaled_variance
            scaled_variances[phase_state] = new_scaled_variance

        # do it once
        reversed_trajectory = trajectory[::-1, :]

        # Welford's online algorithm for mean and variance calculation. See Knuth Vol 2, pg 232
        data_count[:trajectory_len] += 1
        old_mean = np.array(mean[:trajectory_len])  # copy
        mean[:trajectory_len] += \
            (reversed_trajectory - old_mean) / np.expand_dims(data_count[:trajectory_len], axis=-1)
        scaled_variance[:trajectory_len] += (reversed_trajectory - old_mean) * \
                                            (reversed_trajectory - mean[:trajectory_len])

    # need to create a new tuple, as means and scaled variances can get clobbered.
    return TrajectoryStatistics(
            trajectory_length_counts=trajectory_length_counts,
            data_counts=data_counts,
            means=means,
            variances=scaled_variances,
            )


def reducer(init_stats: TrajectoryStatistics,
            new_stats: TrajectoryStatistics) -> TrajectoryStatistics:
    # unpack
    trajectory_length_counts: Dict[HashableNdArray, BinCounter]
    data_counts: Dict[HashableNdArray, np.ndarray]
    means: Dict[HashableNdArray, np.ndarray]
    variances: Dict[HashableNdArray, np.ndarray]
    trajectory_length_counts, data_counts, means, variances = init_stats

    # unpack
    next_trajectory_length_counts: Dict[HashableNdArray, BinCounter]
    next_data_counts: Dict[HashableNdArray, np.ndarray]
    next_means: Dict[HashableNdArray, np.ndarray]
    next_scaled_variances: Dict[HashableNdArray, np.ndarray]
    (
        next_trajectory_length_counts,
        next_data_counts,
        next_means,
        next_scaled_variances,
        ) = new_stats

    for phase_point in next_trajectory_length_counts.keys():
        num_variables = next_means[phase_point].shape[1]
        if phase_point not in trajectory_length_counts:
            # just copy it
            trajectory_length_counts[phase_point] = next_trajectory_length_counts[phase_point]
            data_counts[phase_point] = next_data_counts[phase_point]
            means[phase_point] = next_means[phase_point]
            # remember to scale!
            variances[phase_point] = next_scaled_variances[
                                         phase_point
                                     ] / np.expand_dims(data_counts[phase_point], axis=-1)
        else:
            # merge
            trajectory_length_counts[phase_point].add(
                    next_trajectory_length_counts[phase_point]
                    )
            new_max_len = trajectory_length_counts[phase_point].max

            # data counts
            old_data_count: np.ndarray = data_counts[phase_point]
            next_data_count: np.ndarray = next_data_counts[phase_point]
            new_data_counts = np.zeros(shape=new_max_len, dtype=int)
            new_data_counts[: old_data_count.shape[0]] = old_data_count
            new_data_counts[: next_data_count.shape[0]] += next_data_count

            data_counts[phase_point] = new_data_counts

            # means
            old_mean: np.ndarray = means[phase_point]
            next_mean: np.ndarray = next_means[phase_point]
            new_mean: np.ndarray = np.zeros(
                    shape=(new_max_len, num_variables), dtype=np.float64
                    )
            max_overlap = min(old_mean.shape[0], next_mean.shape[0])
            # weighted average on the overlap, copy over on the tail
            new_mean[:max_overlap] = (
                                             np.expand_dims(old_data_count[:max_overlap], axis=-1)
                                             * old_mean[:max_overlap]
                                             + np.expand_dims(next_data_count[:max_overlap], axis=-1)
                                             * next_mean[:max_overlap]
                                     ) / np.expand_dims(new_data_counts[:max_overlap], axis=-1)
            if old_mean.shape[0] > next_mean.shape[0]:
                new_mean[max_overlap:, :] = old_mean[max_overlap:, :]
            else:
                new_mean[max_overlap:, :] = next_mean[max_overlap:, :]

            means[phase_point] = new_mean

            # scaled variances
            old_variance: np.ndarray = variances[phase_point]
            next_scaled_variance: np.ndarray = next_scaled_variances[phase_point]
            new_variance: np.ndarray = np.zeros(
                    shape=(new_max_len, num_variables), dtype=np.float64
                    )
            # "weighted average" on the overlap, copy over on the tail
            old_proportion = np.expand_dims(
                    old_data_count[:max_overlap] / new_data_counts[:max_overlap], axis=-1
                    )
            next_proportion = np.expand_dims(
                    next_data_count[:max_overlap] / new_data_counts[:max_overlap], axis=-1
                    )
            new_variance[:max_overlap] = (
                    old_proportion * old_variance[:max_overlap]
                    + next_proportion
                    * (next_scaled_variance[:max_overlap] / np.expand_dims(next_data_count[:max_overlap], axis=-1))
                    + ((old_mean[:max_overlap] - next_mean[:max_overlap]) ** 2)
                    * old_proportion
                    * next_proportion)
            if old_variance.shape[0] > next_scaled_variance.shape[0]:
                new_variance[max_overlap:, :] = old_variance[max_overlap:, :]
            else:
                new_variance[max_overlap:, :] = next_scaled_variance[
                                                max_overlap:, :
                                                ] / np.expand_dims(next_data_count[max_overlap:], axis=-1)

            variances[phase_point] = new_variance

    return TrajectoryStatistics(trajectory_length_counts, data_counts, means, variances)


def compute_cycles(*,
                   model_text: str,
                   knockouts: Dict[str, int],
                   continuous: Dict[str, bool],
                   num_iterations: int,
                   visualize_variables: Dict[str, bool]):
    """

    Parameters
    ----------
    model_text
    knockouts
    continuous
    num_iterations
    visualize_variables

    Returns
    -------

    """
    # turn boolean-valued dictionary to list of variables to visualize
    visualized_variables: Tuple[str] = tuple(
            var for var in visualize_variables if visualize_variables[var]
            )

    # create an update function and equation system
    variables, update_fn, equation_system = process_model_text(
            model_text, knockouts, continuous
            )

    # compile and warm up the jitter
    update_fn = numba.jit(update_fn)
    update_fn(np.zeros(len(variables), dtype=np.int64))

    # associate variable names with their index in vectors
    variable_idx: Dict[str, int] = dict(zip(variables, range(len(variables))))

    constants: List[str] = equation_system.constant_variables()
    constants_vals: Dict[str, int] = {
        const: int(equation_system[const]) for const in constants
        }

    # decide if we will perform a complete state space search or not
    state_space_size = 3 ** (len(variables) - len(constants))
    perform_complete_search = state_space_size <= num_iterations

    state_generator: Iterator[np.ndarray]
    if perform_complete_search:
        state_generator = complete_search_generator(
                variables=variables, constants_vals=constants_vals
                )
        num_iterations = state_space_size
    else:
        state_generator = random_search_generator(
                num_iterations=num_iterations,
                variables=variables,
                constants_vals=constants_vals,
                )

    max_threads = max(1, min(pathos.multiprocessing.cpu_count() - 1, math.floor(state_space_size / 1000)))

    batch_generator = batcher(
            state_generator, variables, batch_size=min(1000, num_iterations // max_threads)
            )

    trajectory_length_counts: Dict[HashableNdArray, BinCounter] = dict()
    data_counts: Dict[HashableNdArray, np.ndarray] = dict()
    means: Dict[HashableNdArray, np.ndarray] = dict()
    variances: Dict[HashableNdArray, np.ndarray] = dict()
    with pathos.multiprocessing.ProcessPool(nodes=max_threads) as pool:
        trajectory_length_counts, data_counts, means, variances = functools.reduce(
                reducer,
                pool.uimap(
                        batch_trajectory_process, batch_generator, itertools.repeat(update_fn)
                        ),
                TrajectoryStatistics(
                        trajectory_length_counts, data_counts, means, variances
                        ),
                )

    phase_points = trajectory_length_counts.keys()

    # recreate limit cycles from phase point keys
    limit_cycles: Dict[HashableNdArray, List] = dict()  # record limit sets with their ordering. i.e. as cycles
    for phase_point in phase_points:
        cycle = list()
        t_state = phase_point
        state = phase_point.array
        # do-while
        cycle.append(t_state)
        state = update_fn(state)
        t_state = HashableNdArray(state)
        while t_state != phase_point:
            cycle.append(t_state)
            state = update_fn(state)
            t_state = HashableNdArray(state)
        # record cycle
        limit_cycles[phase_point] = cycle

    # create images for each variable
    limit_set_stats_images = {phase_point: dict() for phase_point in phase_points}
    with tempfile.TemporaryDirectory() as tmp_dir_name:
        plt.rcParams["svg.fonttype"] = "none"
        for phase_point, var in itertools.product(phase_points, visualized_variables):
            cycle = limit_cycles[phase_point]
            var_idx = variable_idx[var]

            var_means = np.flipud(means[phase_point][:, var_idx])
            var_stdevs = np.flipud(np.sqrt(variances[phase_point][:, var_idx]))

            plt.figure(figsize=(6, 4))

            # means, since phased will fix on single value at end
            plt.plot(list(var_means) + [cycle[0][var_idx]], color="#1f3d87")
            plt.fill_between(
                    range(len(var_means) + 1),
                    list(var_means - var_stdevs) + [cycle[0][var_idx]],
                    list(var_means + var_stdevs) + [cycle[0][var_idx]],
                    color="grey",
                    alpha=0.25,
                    )
            # dashed cycle in `converging` region
            plt.plot(
                    range(len(var_means) - len(cycle), len(var_means) + 1),
                    [cycle[idx][var_idx] for idx in range(len(cycle))]
                    + [cycle[0][var_idx]],
                    color="#F24F00",
                    linestyle="dashed",
                    )
            # solid cycle in `converged` region
            plt.plot(
                    range(len(var_means), len(var_means) + len(cycle) + 1),
                    [cycle[idx][var_idx] for idx in range(len(cycle))]
                    + [cycle[0][var_idx]],
                    color="#F24F00",
                    )

            plt.title(f"Phased Envelope of trajectories for {var}")

            plt.xlabel("Distance to phase-fixing point on limit cycle")
            max_tick_goal = 40
            tick_gap = max(1, ceil((len(var_means) + len(cycle) + 1) / max_tick_goal))
            x_tick_locations = [
                x
                for x in range(0, len(var_means) + len(cycle) + 1)
                if (x - len(var_means)) % tick_gap == 0
                ]
            x_tick_labels = list(map(lambda x: len(var_means) - x, x_tick_locations))
            plt.xticks(x_tick_locations, x_tick_labels, rotation=90)
            plt.xlim([0, len(var_means) + len(cycle)])
            plt.axvline(len(var_means), linestyle="dotted", color="grey")
            plt.axvline(len(var_means) - len(cycle), linestyle="dotted", color="grey")

            plt.ylabel("State")
            plt.yticks([0, 1, 2])
            for y_val in range(3):
                plt.axhline(y=y_val, linestyle="dotted", color="grey", alpha=0.5)
            plt.ylim([0 - 0.1, 2 + 0.1])

            plt.tight_layout()

            image_filename = "dist-" + str(hash(phase_point)) + "-" + var + ".svg"
            plt.savefig(
                    tmp_dir_name + "/" + image_filename, transparent=True, pad_inches=0.0
                    )
            plt.close()
            with open(tmp_dir_name + "/" + image_filename, "r") as image:
                limit_set_stats_images[phase_point][var] = Markup(image.read())

    # print('var images complete')

    # cycle list sorted by frequency
    cycle_list = list(limit_cycles.values())
    cycle_list.sort(
            key=lambda cyc: trajectory_length_counts[cyc[0]].total(), reverse=True
            )

    # tally of number of cycles of each length, sorted by cycle length
    cycle_len_counts = defaultdict(lambda: 0)
    for cycle in cycle_list:
        cycle_len_counts[len(cycle)] += 1
    cycle_len_counts = tuple(sorted(cycle_len_counts.items()))

    length_distribution_images = dict()
    with tempfile.TemporaryDirectory() as tmp_dir_name:
        # create length distribution plots
        plt.rcParams["svg.fonttype"] = "none"
        for phase_point in phase_points:
            length_dist = trajectory_length_counts[phase_point].trimmed_bins()
            plt.figure(figsize=(4, 3))
            plt.bar(x=range(len(length_dist)), height=length_dist, color="#002868")
            plt.title("Distribution of lengths of paths")
            plt.xlabel("Length of path")
            plt.ylabel("Number of states")
            plt.tight_layout()
            image_filename = "dist-" + str(hash(phase_point)) + ".svg"
            plt.savefig(
                    tmp_dir_name + "/" + image_filename, transparent=True, pad_inches=0.0
                    )
            plt.close()
            with open(tmp_dir_name + "/" + image_filename, "r") as image:
                length_distribution_images[phase_point] = Markup(image.read())

    def get_key(cyc):
        return cyc[0]

    def get_count(cyc):
        return trajectory_length_counts[get_key(cyc)].total()

    cycles = [
        {
            "states": cycle,
            "len": len(cycle),
            "count": get_count(cycle),
            "percent": 100 * get_count(cycle) / num_iterations,
            "len-dist-image": length_distribution_images[get_key(cycle)]
            if get_key(cycle) in length_distribution_images
            else "",
            "limit-set-stats-images": limit_set_stats_images[get_key(cycle)]
            if get_key(cycle) in limit_set_stats_images
            else dict(),
            }
        for cycle in cycle_list]

    # respond with the results-of-computation page
    return make_response(
            render_template(
                    "compute-cycles.html",
                    cycles=cycles,
                    variables=variables,
                    cycle_len_counts=cycle_len_counts,
                    complete_results=perform_complete_search,
                    )
            )
